@@ -3,49 +3,31 @@
  *
  * Copyright (C) 2010-2015  Tetsuo Handa <penguin-kernel@I-love.SAKURA.ne.jp>
  *
- * Version: 1.0.38   2018/04/01
+ * Version: 1.0.39   2019/08/20
  */
 
 #include "internal.h"
 #include "probe.h"
 
 /* Prototype definition. */
-static void ccs_task_security_gc(void);
-static int ccs_copy_cred_security(const struct cred *new,
-				  const struct cred *old, gfp_t gfp);
-static struct ccs_security *ccs_find_cred_security(const struct cred *cred);
-static DEFINE_SPINLOCK(ccs_task_security_list_lock);
-static atomic_t ccs_in_execve_tasks = ATOMIC_INIT(0);
-/*
- * List of "struct ccs_security" for "struct pid".
- *
- * All instances on this list is guaranteed that "struct ccs_security"->pid !=
- * NULL. Also, instances on this list that are in execve() are guaranteed that
- * "struct ccs_security"->cred remembers "struct linux_binprm"->cred with a
- * refcount on "struct linux_binprm"->cred.
- */
-struct list_head ccs_task_security_list[CCS_MAX_TASK_SECURITY_HASH];
-/*
- * List of "struct ccs_security" for "struct cred".
- *
- * Since the number of "struct cred" is nearly equals to the number of
- * "struct pid", we allocate hash tables like ccs_task_security_list.
- *
- * All instances on this list are guaranteed that "struct ccs_security"->pid ==
- * NULL and "struct ccs_security"->cred != NULL.
- */
-static struct list_head ccs_cred_security_list[CCS_MAX_TASK_SECURITY_HASH];
+static int __ccs_alloc_task_security(const struct task_struct *task);
+static void __ccs_free_task_security(const struct task_struct *task);
 
 /* Dummy security context for avoiding NULL pointer dereference. */
 static struct ccs_security ccs_oom_security = {
-	.ccs_domain_info = &ccs_kernel_domain
+        .ccs_domain_info = &ccs_kernel_domain
 };
 
 /* Dummy security context for avoiding NULL pointer dereference. */
 static struct ccs_security ccs_default_security = {
-	.ccs_domain_info = &ccs_kernel_domain
+        .ccs_domain_info = &ccs_kernel_domain
 };
 
+/* List of "struct ccs_security". */
+struct list_head ccs_task_security_list[CCS_MAX_TASK_SECURITY_HASH];
+/* Lock for protecting ccs_task_security_list[]. */
+static DEFINE_SPINLOCK(ccs_task_security_list_lock);
+
 /* For exporting variables and functions. */
 struct ccsecurity_exports ccsecurity_exports;
 /* Members are updated by loadable kernel module. */
@@ -53,8 +35,8 @@ struct ccsecurity_operations ccsecurity_ops;
 
 /* Original hooks. */
 static union security_list_options original_cred_prepare;
-static union security_list_options original_cred_free;
-static union security_list_options original_cred_alloc_blank;
+static union security_list_options original_task_alloc;
+static union security_list_options original_task_free;
 
 #ifdef CONFIG_AKARI_TRACE_EXECVE_COUNT
 
@@ -115,7 +97,7 @@ void ccs_audit_free_execve(const struct ccs_execve * const ee,
 		static bool done;					\
 		if (!done) {						\
 			printk(KERN_INFO				\
-			       "AKARI: Debug trace: " pos " of 4\n");	\
+			       "AKARI: Debug trace: " pos " of 2\n");	\
 			done = true;					\
 		}							\
 	} while (0)
@@ -131,115 +113,74 @@ void ccs_audit_free_execve(const struct ccs_execve * const ee,
  */
 static void ccs_clear_execve(int ret, struct ccs_security *security)
 {
-	struct ccs_execve *ee;
-	if (security == &ccs_default_security || security == &ccs_oom_security)
+	struct ccs_execve *ee = security->ee;
+	if (security == &ccs_default_security || security == &ccs_oom_security
+	    || !ee)
 		return;
-	ee = security->ee;
 	security->ee = NULL;
-	if (!ee)
-		return;
-	atomic_dec(&ccs_in_execve_tasks);
 	ccs_finish_execve(ret, ee);
 }
 
 /**
- * ccs_rcu_free - RCU callback for releasing "struct ccs_security".
- *
- * @rcu: Pointer to "struct rcu_head".
- *
- * Returns nothing.
- */
-static void ccs_rcu_free(struct rcu_head *rcu)
-{
-	struct ccs_security *ptr = container_of(rcu, typeof(*ptr), rcu);
-	struct ccs_execve *ee = ptr->ee;
-	/*
-	 * If this security context was associated with "struct pid" and
-	 * ptr->ccs_flags has CCS_TASK_IS_IN_EXECVE set, it indicates that a
-	 * "struct task_struct" associated with this security context exited
-	 * immediately after do_execve() has failed.
-	 */
-	if (ptr->pid && (ptr->ccs_flags & CCS_TASK_IS_IN_EXECVE)) {
-		ccs_debug_trace("1");
-		atomic_dec(&ccs_in_execve_tasks);
-	}
-	/*
-	 * If this security context was associated with "struct pid",
-	 * drop refcount obtained by get_pid() in ccs_find_task_security().
-	 */
-	if (ptr->pid) {
-		ccs_debug_trace("2");
-		put_pid(ptr->pid);
-	}
-	if (ee) {
-		ccs_debug_trace("3");
-		ccs_audit_free_execve(ee, false);
-		kfree(ee->handler_path);
-		kfree(ee);
-	}
-	kfree(ptr);
-}
-
-/**
- * ccs_del_security - Release "struct ccs_security".
- *
- * @ptr: Pointer to "struct ccs_security".
- *
- * Returns nothing.
- */
-static void ccs_del_security(struct ccs_security *ptr)
-{
-	unsigned long flags;
-	if (ptr == &ccs_default_security || ptr == &ccs_oom_security)
-		return;
-	spin_lock_irqsave(&ccs_task_security_list_lock, flags);
-	list_del_rcu(&ptr->list);
-	spin_unlock_irqrestore(&ccs_task_security_list_lock, flags);
-	call_rcu(&ptr->rcu, ccs_rcu_free);
-}
-
-/**
- * ccs_add_cred_security - Add "struct ccs_security" to list.
- *
- * @ptr: Pointer to "struct ccs_security".
- *
- * Returns nothing.
- */
-static void ccs_add_cred_security(struct ccs_security *ptr)
-{
-	unsigned long flags;
-	struct list_head *list = &ccs_cred_security_list
-		[hash_ptr((void *) ptr->cred, CCS_TASK_SECURITY_HASH_BITS)];
-#ifdef CONFIG_AKARI_DEBUG
-	if (ptr->pid)
-		printk(KERN_INFO "AKARI: \"struct ccs_security\"->pid != NULL"
-		       "\n");
-#endif
-	ptr->pid = NULL;
-	spin_lock_irqsave(&ccs_task_security_list_lock, flags);
-	list_add_rcu(&ptr->list, list);
-	spin_unlock_irqrestore(&ccs_task_security_list_lock, flags);
-}
-
-/**
- * ccs_task_create - Make snapshot of security context for new task.
+ * ccs_task_alloc_security - Allocate memory for new tasks.
  *
+ * @p: Pointer to "struct task_struct".
  * @clone_flags: Flags passed to clone().
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_task_create(unsigned long clone_flags)
+static int ccs_task_alloc_security(struct task_struct *p,
+				   unsigned long clone_flags)
 {
-	struct ccs_security *old_security;
-	struct ccs_security *new_security;
-	struct cred *cred = prepare_creds();
-	if (!cred)
-		return -ENOMEM;
-	old_security = ccs_find_task_security(current);
-	new_security = ccs_find_cred_security(cred);
-	new_security->ccs_domain_info = old_security->ccs_domain_info;
-	new_security->ccs_flags = old_security->ccs_flags;
-	return commit_creds(cred);
+	int rc = __ccs_alloc_task_security(p);
+	if (rc)
+		return rc;
+	if (original_task_alloc.task_alloc) {
+		rc = original_task_alloc.task_alloc(p, clone_flags);
+		if (rc)
+			__ccs_free_task_security(p);
+	}
+	return rc;
+}
+
+/**
+ * ccs_task_free_security - Release memory for "struct task_struct".
+ *
+ * @p: Pointer to "struct task_struct".
+ *
+ * Returns nothing.
+ */
+static void ccs_task_free_security(struct task_struct *p)
+{
+	struct ccs_security *ptr = ccs_find_task_security(p);
+	struct ccs_execve *ee = ptr->ee;
+	if (original_task_free.task_free)
+		original_task_free.task_free(p);
+	/*
+	 * Since an LSM hook for reverting domain transition is missing,
+	 * ccs_finish_execve() is not called if exited immediately after
+	 * execve() failed.
+	 */
+	if (ee) {
+		ccs_debug_trace("2");
+		ccs_audit_free_execve(ee, false);
+		kfree(ee->handler_path);
+		kfree(ee);
+		ptr->ee = NULL;
+	}
+	__ccs_free_task_security(p);
+}
+
+/**
+ * ccs_bprm_committing_creds - A hook which is called when do_execve() succeeded.
+ *
+ * @bprm: Pointer to "struct linux_binprm".
+ *
+ * Returns nothing.
+ */
+static void ccs_bprm_committing_creds(struct linux_binprm *bprm)
+{
+	ccs_clear_execve(0, ccs_current_security());
 }
 
 /**
@@ -254,7 +195,6 @@ static int ccs_task_create(unsigned long clone_flags)
 static int ccs_cred_prepare(struct cred *new, const struct cred *old,
 			    gfp_t gfp)
 {
-	int rc1;
 	/*
 	 * For checking whether reverting domain transition is needed or not.
 	 *
@@ -262,117 +202,11 @@ static int ccs_cred_prepare(struct cred *new, const struct cred *old,
 	 */
 	if (gfp == GFP_KERNEL)
 		ccs_find_task_security(current);
-	rc1 = ccs_copy_cred_security(new, old, gfp);
-	if (gfp == GFP_KERNEL)
-		ccs_task_security_gc();
-	if (original_cred_prepare.cred_prepare) {
-		const int rc2 = original_cred_prepare.cred_prepare(new, old,
-								   gfp);
-		if (rc2) {
-			ccs_del_security(ccs_find_cred_security(new));
-			return rc2;
-		}
-	}
-	return rc1;
-}
-
-/**
- * ccs_cred_free - Release memory used by credentials.
- *
- * @cred: Pointer to "struct cred".
- *
- * Returns nothing.
- */
-static void ccs_cred_free(struct cred *cred)
-{
-	if (original_cred_free.cred_free)
-		original_cred_free.cred_free(cred);
-	ccs_del_security(ccs_find_cred_security(cred));
-}
-
-/**
- * ccs_alloc_cred_security - Allocate memory for new credentials.
- *
- * @cred: Pointer to "struct cred".
- * @gfp:  Memory allocation flags.
- *
- * Returns 0 on success, negative value otherwise.
- */
-static int ccs_alloc_cred_security(const struct cred *cred, gfp_t gfp)
-{
-	struct ccs_security *new_security = kzalloc(sizeof(*new_security),
-						    gfp);
-	if (!new_security)
-		return -ENOMEM;
-	new_security->cred = cred;
-	ccs_add_cred_security(new_security);
+	if (original_cred_prepare.cred_prepare)
+		return original_cred_prepare.cred_prepare(new, old, gfp);
 	return 0;
 }
 
-/**
- * ccs_cred_alloc_blank - Allocate memory for new credentials.
- *
- * @new: Pointer to "struct cred".
- * @gfp: Memory allocation flags.
- *
- * Returns 0 on success, negative value otherwise.
- */
-static int ccs_cred_alloc_blank(struct cred *new, gfp_t gfp)
-{
-	const int rc1 = ccs_alloc_cred_security(new, gfp);
-	if (original_cred_alloc_blank.cred_alloc_blank) {
-		const int rc2 = original_cred_alloc_blank.
-			cred_alloc_blank(new, gfp);
-		if (rc2) {
-			ccs_del_security(ccs_find_cred_security(new));
-			return rc2;
-		}
-	}
-	return rc1;
-}
-
-/**
- * ccs_cred_transfer - Transfer "struct ccs_security" between credentials.
- *
- * @new: Pointer to "struct cred".
- * @old: Pointer to "struct cred".
- *
- * Returns nothing.
- */
-static void ccs_cred_transfer(struct cred *new, const struct cred *old)
-{
-	struct ccs_security *new_security = ccs_find_cred_security(new);
-	struct ccs_security *old_security = ccs_find_cred_security(old);
-	if (new_security == &ccs_default_security ||
-	    new_security == &ccs_oom_security ||
-	    old_security == &ccs_default_security ||
-	    old_security == &ccs_oom_security)
-		return;
-	new_security->ccs_flags = old_security->ccs_flags;
-	new_security->ccs_domain_info = old_security->ccs_domain_info;
-}
-
-/**
- * ccs_bprm_committing_creds - A hook which is called when do_execve() succeeded.
- *
- * @bprm: Pointer to "struct linux_binprm".
- *
- * Returns nothing.
- */
-static void ccs_bprm_committing_creds(struct linux_binprm *bprm)
-{
-	struct ccs_security *old_security = ccs_current_security();
-	struct ccs_security *new_security;
-	if (old_security == &ccs_default_security ||
-	    old_security == &ccs_oom_security)
-		return;
-	ccs_clear_execve(0, old_security);
-	/* Update current task's cred's domain for future fork(). */
-	new_security = ccs_find_cred_security(bprm->cred);
-	new_security->ccs_flags = old_security->ccs_flags;
-	new_security->ccs_domain_info = old_security->ccs_domain_info;
-}
-
 /**
  * ccs_bprm_check_security - Check permission for execve().
  *
@@ -383,7 +217,6 @@ static void ccs_bprm_committing_creds(struct linux_binprm *bprm)
 static int ccs_bprm_check_security(struct linux_binprm *bprm)
 {
 	struct ccs_security *security = ccs_current_security();
-	int rc;
 	if (security == &ccs_default_security || security == &ccs_oom_security)
 		return -ENOMEM;
 	if (security->ee)
@@ -392,12 +225,22 @@ static int ccs_bprm_check_security(struct linux_binprm *bprm)
 	if (!ccs_policy_loaded)
 		ccs_load_policy(bprm->filename);
 #endif
-	rc = ccs_start_execve(bprm, &security->ee);
-	if (security->ee)
-		atomic_inc(&ccs_in_execve_tasks);
-	return rc;
+	return ccs_start_execve(bprm, &security->ee);
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)
+/**
+ * ccs_file_open - Check permission for open().
+ *
+ * @f:    Pointer to "struct file".
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_file_open(struct file *f)
+{
+	return ccs_open_permission(f);
+}
+#else
 /**
  * ccs_file_open - Check permission for open().
  *
@@ -410,6 +253,7 @@ static int ccs_file_open(struct file *f, const struct cred *cred)
 {
 	return ccs_open_permission(f);
 }
+#endif
 
 #ifdef CONFIG_SECURITY_PATH
 
@@ -422,7 +266,7 @@ static int ccs_file_open(struct file *f, const struct cred *cred)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_chown(struct path *path, kuid_t user, kgid_t group)
+static int ccs_path_chown(const struct path *path, kuid_t user, kgid_t group)
 {
 	return ccs_chown_permission(path->dentry, path->mnt, user, group);
 }
@@ -435,7 +279,7 @@ static int ccs_path_chown(struct path *path, kuid_t user, kgid_t group)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_chmod(struct path *path, umode_t mode)
+static int ccs_path_chmod(const struct path *path, umode_t mode)
 {
 	return ccs_chmod_permission(path->dentry, path->mnt, mode);
 }
@@ -447,7 +291,7 @@ static int ccs_path_chmod(struct path *path, umode_t mode)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_chroot(struct path *path)
+static int ccs_path_chroot(const struct path *path)
 {
 	return ccs_chroot_permission(path);
 }
@@ -459,7 +303,7 @@ static int ccs_path_chroot(struct path *path)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_truncate(struct path *path)
+static int ccs_path_truncate(const struct path *path)
 {
 	return ccs_truncate_permission(path->dentry, path->mnt);
 }
@@ -521,7 +365,7 @@ static int ccs_inode_getattr(const struct path *path)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_mknod(struct path *dir, struct dentry *dentry,
+static int ccs_path_mknod(const struct path *dir, struct dentry *dentry,
 			  umode_t mode, unsigned int dev)
 {
 	return ccs_mknod_permission(dentry, dir->mnt, mode, dev);
@@ -536,7 +380,7 @@ static int ccs_path_mknod(struct path *dir, struct dentry *dentry,
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_mkdir(struct path *dir, struct dentry *dentry,
+static int ccs_path_mkdir(const struct path *dir, struct dentry *dentry,
 			  umode_t mode)
 {
 	return ccs_mkdir_permission(dentry, dir->mnt, mode);
@@ -550,7 +394,7 @@ static int ccs_path_mkdir(struct path *dir, struct dentry *dentry,
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_rmdir(struct path *dir, struct dentry *dentry)
+static int ccs_path_rmdir(const struct path *dir, struct dentry *dentry)
 {
 	return ccs_rmdir_permission(dentry, dir->mnt);
 }
@@ -563,7 +407,7 @@ static int ccs_path_rmdir(struct path *dir, struct dentry *dentry)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_unlink(struct path *dir, struct dentry *dentry)
+static int ccs_path_unlink(const struct path *dir, struct dentry *dentry)
 {
 	return ccs_unlink_permission(dentry, dir->mnt);
 }
@@ -577,7 +421,7 @@ static int ccs_path_unlink(struct path *dir, struct dentry *dentry)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_symlink(struct path *dir, struct dentry *dentry,
+static int ccs_path_symlink(const struct path *dir, struct dentry *dentry,
 			    const char *old_name)
 {
 	return ccs_symlink_permission(dentry, dir->mnt, old_name);
@@ -593,8 +437,10 @@ static int ccs_path_symlink(struct path *dir, struct dentry *dentry,
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_rename(struct path *old_dir, struct dentry *old_dentry,
-			   struct path *new_dir, struct dentry *new_dentry)
+static int ccs_path_rename(const struct path *old_dir,
+			   struct dentry *old_dentry,
+			   const struct path *new_dir,
+			   struct dentry *new_dentry)
 {
 	return ccs_rename_permission(old_dentry, new_dentry, old_dir->mnt);
 }
@@ -608,7 +454,7 @@ static int ccs_path_rename(struct path *old_dir, struct dentry *old_dentry,
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_path_link(struct dentry *old_dentry, struct path *new_dir,
+static int ccs_path_link(struct dentry *old_dentry, const struct path *new_dir,
 			 struct dentry *new_dentry)
 {
 	return ccs_link_permission(old_dentry, new_dentry, new_dir->mnt);
@@ -1059,7 +905,8 @@ static void ccs_inode_free_security(struct inode *inode)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_sb_pivotroot(struct path *old_path, struct path *new_path)
+static int ccs_sb_pivotroot(const struct path *old_path,
+			    const struct path *new_path)
 {
 	return ccs_pivot_root_permission(old_path, new_path);
 }
@@ -1075,12 +922,27 @@ static int ccs_sb_pivotroot(struct path *old_path, struct path *new_path)
  *
  * Returns 0 on success, negative value otherwise.
  */
-static int ccs_sb_mount(const char *dev_name, struct path *path,
+static int ccs_sb_mount(const char *dev_name, const struct path *path,
 			const char *type, unsigned long flags, void *data_page)
 {
 	return ccs_mount_permission(dev_name, path, type, flags, data_page);
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 2, 0)
+/**
+ * ccs_move_mount - Check permission for move_mount() operation.
+ *
+ * @from_path: Pointer to "struct path".
+ * @to_path:   Pointer to "struct path".
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_move_mount(const struct path *from_path, const struct path *to_path)
+{
+	return ccs_move_mount_permission(from_path, to_path);
+}
+#endif
+
 /**
  * ccs_sb_umount - Check permission for umount().
  *
@@ -1130,11 +992,9 @@ static int ccs_file_ioctl(struct file *filp, unsigned int cmd,
 
 static struct security_hook_list akari_hooks[] = {
 	/* Security context allocator. */
-	MY_HOOK_INIT(cred_free, ccs_cred_free),
+	MY_HOOK_INIT(task_free, ccs_task_free_security),
 	MY_HOOK_INIT(cred_prepare, ccs_cred_prepare),
-	MY_HOOK_INIT(cred_alloc_blank, ccs_cred_alloc_blank),
-	MY_HOOK_INIT(cred_transfer, ccs_cred_transfer),
-	MY_HOOK_INIT(task_create, ccs_task_create),
+	MY_HOOK_INIT(task_alloc, ccs_task_alloc_security),
 	/* Security context updater for successful execve(). */
 	MY_HOOK_INIT(bprm_check_security, ccs_bprm_check_security),
 	MY_HOOK_INIT(bprm_committing_creds, ccs_bprm_committing_creds),
@@ -1144,6 +1004,9 @@ static struct security_hook_list akari_hooks[] = {
 	MY_HOOK_INIT(file_ioctl, ccs_file_ioctl),
 	MY_HOOK_INIT(sb_pivotroot, ccs_sb_pivotroot),
 	MY_HOOK_INIT(sb_mount, ccs_sb_mount),
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 2, 0)
+	MY_HOOK_INIT(move_mount, ccs_move_mount),
+#endif
 	MY_HOOK_INIT(sb_umount, ccs_sb_umount),
 #ifdef CONFIG_SECURITY_PATH
 	MY_HOOK_INIT(path_mknod, ccs_path_mknod),
@@ -1187,13 +1050,34 @@ static struct security_hook_list akari_hooks[] = {
 
 static inline void add_hook(struct security_hook_list *hook)
 {
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 17, 0)
+	hlist_add_tail_rcu(&hook->list, hook->head);
+#else
 	list_add_tail_rcu(&hook->list, hook->head);
+#endif
 }
 
 static void __init swap_hook(struct security_hook_list *hook,
 			     union security_list_options *original)
 {
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 17, 0)
+	struct hlist_head *list = hook->head;
+
+	if (hlist_empty(list)) {
+		add_hook(hook);
+	} else {
+		struct security_hook_list *shp =
+			hlist_entry(list->first, typeof(*shp), list);
+
+		while (shp->list.next)
+			shp = hlist_entry(shp->list.next, typeof(*shp), list);
+		*original = shp->hook;
+		smp_wmb();
+		shp->hook = hook->hook;
+	}
+#else
 	struct list_head *list = hook->head;
+
 	if (list_empty(list)) {
 		add_hook(hook);
 	} else {
@@ -1203,8 +1087,91 @@ static void __init swap_hook(struct security_hook_list *hook,
 		smp_wmb();
 		shp->hook = hook->hook;
 	}
+#endif
 }
 
+#if defined(CONFIG_STRICT_KERNEL_RWX) && !defined(CONFIG_SECURITY_WRITABLE_HOOKS)
+#include <linux/uaccess.h> /* probe_kernel_write() */
+#define NEED_TO_CHECK_HOOKS_ARE_WRITABLE
+
+#if defined(CONFIG_X86)
+#define MAX_RO_PAGES 1024
+static struct page *ro_pages[MAX_RO_PAGES] __initdata;
+static unsigned int ro_pages_len __initdata;
+
+static bool __init lsm_test_page_ro(void *addr)
+{
+	unsigned int i;
+	int unused;
+	struct page *page;
+
+	page = (struct page *) lookup_address((unsigned long) addr, &unused);
+	if (!page)
+		return false;
+	if (test_bit(_PAGE_BIT_RW, &(page->flags)))
+		return true;
+	for (i = 0; i < ro_pages_len; i++)
+		if (page == ro_pages[i])
+			return true;
+	if (ro_pages_len == MAX_RO_PAGES)
+		return false;
+	ro_pages[ro_pages_len++] = page;
+	return true;
+}
+
+static bool __init check_ro_pages(struct security_hook_heads *hooks)
+{
+	int i;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 17, 0)
+	struct hlist_head *list = &hooks->capable;
+
+	if (!probe_kernel_write(list, list, sizeof(void *)))
+		return true;
+	for (i = 0; i < ARRAY_SIZE(akari_hooks); i++) {
+		struct hlist_head *head = akari_hooks[i].head;
+		struct security_hook_list *shp;
+
+		if (!lsm_test_page_ro(&head->first))
+			return false;
+		hlist_for_each_entry(shp, head, list)
+			if (!lsm_test_page_ro(&shp->list.next) ||
+			    !lsm_test_page_ro(&shp->list.pprev))
+				return false;
+	}
+#else
+	struct list_head *list = &hooks->capable;
+
+	if (!probe_kernel_write(list, list, sizeof(void *)))
+		return true;
+	for (i = 0; i < ARRAY_SIZE(akari_hooks); i++) {
+		struct list_head *head = akari_hooks[i].head;
+		struct security_hook_list *shp;
+
+		if (!lsm_test_page_ro(&head->next) ||
+		    !lsm_test_page_ro(&head->prev))
+			return false;
+		list_for_each_entry(shp, head, list)
+			if (!lsm_test_page_ro(&shp->list.next) ||
+			    !lsm_test_page_ro(&shp->list.prev))
+				return false;
+	}
+#endif
+	return true;
+}
+#else
+static bool __init check_ro_pages(struct security_hook_heads *hooks)
+{
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 17, 0)
+	struct hlist_head *list = &hooks->capable;
+#else
+	struct list_head *list = &hooks->capable;
+#endif
+
+	return !probe_kernel_write(list, list, sizeof(void *));
+}
+#endif
+#endif
+
 /**
  * ccs_init - Initialize this module.
  *
@@ -1220,6 +1187,12 @@ static int __init ccs_init(void)
 		akari_hooks[idx].head = ((void *) hooks)
 			+ ((unsigned long) akari_hooks[idx].head)
 			- ((unsigned long) &probe_dummy_security_hook_heads);
+#if defined(NEED_TO_CHECK_HOOKS_ARE_WRITABLE)
+	if (!check_ro_pages(hooks)) {
+		printk(KERN_INFO "Can't update security_hook_heads due to write protected. Retry with rodata=0 kernel command line option added.\n");
+		return -EINVAL;
+	}
+#endif
 	ccsecurity_exports.find_task_by_vpid = probe_find_task_by_vpid();
 	if (!ccsecurity_exports.find_task_by_vpid)
 		goto out;
@@ -1229,17 +1202,23 @@ static int __init ccs_init(void)
 	ccsecurity_exports.d_absolute_path = probe_d_absolute_path();
 	if (!ccsecurity_exports.d_absolute_path)
 		goto out;
-	for (idx = 0; idx < CCS_MAX_TASK_SECURITY_HASH; idx++) {
-		INIT_LIST_HEAD(&ccs_cred_security_list[idx]);
+	for (idx = 0; idx < CCS_MAX_TASK_SECURITY_HASH; idx++)
 		INIT_LIST_HEAD(&ccs_task_security_list[idx]);
-	}
 	ccs_main_init();
-	swap_hook(&akari_hooks[0], &original_cred_free);
+#if defined(NEED_TO_CHECK_HOOKS_ARE_WRITABLE) && defined(CONFIG_X86)
+	for (idx = 0; idx < ro_pages_len; idx++)
+		set_bit(_PAGE_BIT_RW, &(ro_pages[idx]->flags));
+#endif
+	swap_hook(&akari_hooks[0], &original_task_free);
 	swap_hook(&akari_hooks[1], &original_cred_prepare);
-	swap_hook(&akari_hooks[2], &original_cred_alloc_blank);
+	swap_hook(&akari_hooks[2], &original_task_alloc);
 	for (idx = 3; idx < ARRAY_SIZE(akari_hooks); idx++)
 		add_hook(&akari_hooks[idx]);
-	printk(KERN_INFO "AKARI: 1.0.38   2018/04/01\n");
+#if defined(NEED_TO_CHECK_HOOKS_ARE_WRITABLE) && defined(CONFIG_X86)
+	for (idx = 0; idx < ro_pages_len; idx++)
+		clear_bit(_PAGE_BIT_RW, &(ro_pages[idx]->flags));
+#endif
+	printk(KERN_INFO "AKARI: 1.0.39   2019/08/20\n");
 	printk(KERN_INFO
 	       "Access Keeping And Regulating Instrument registered.\n");
 	return 0;
@@ -1261,18 +1240,6 @@ MODULE_LICENSE("GPL");
  */
 bool ccs_used_by_cred(const struct ccs_domain_info *domain)
 {
-	int idx;
-	struct ccs_security *ptr;
-	for (idx = 0; idx < CCS_MAX_TASK_SECURITY_HASH; idx++) {
-		struct list_head *list = &ccs_cred_security_list[idx];
-		list_for_each_entry_rcu(ptr, list, list) {
-			struct ccs_execve *ee = ptr->ee;
-			if (ptr->ccs_domain_info == domain ||
-			    (ee && ee->previous_domain == domain)) {
-				return true;
-			}
-		}
-	}
 	return false;
 }
 
@@ -1293,6 +1260,29 @@ static void ccs_add_task_security(struct ccs_security *ptr,
 	spin_unlock_irqrestore(&ccs_task_security_list_lock, flags);
 }
 
+/**
+ * __ccs_alloc_task_security - Allocate memory for new tasks.
+ *
+ * @task: Pointer to "struct task_struct".
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int __ccs_alloc_task_security(const struct task_struct *task)
+{
+	struct ccs_security *old_security = ccs_current_security();
+	struct ccs_security *new_security = kzalloc(sizeof(*new_security),
+						    GFP_KERNEL);
+	struct list_head *list = &ccs_task_security_list
+		[hash_ptr((void *) task, CCS_TASK_SECURITY_HASH_BITS)];
+	if (!new_security)
+		return -ENOMEM;
+	new_security->task = task;
+	new_security->ccs_domain_info = old_security->ccs_domain_info;
+	new_security->ccs_flags = old_security->ccs_flags;
+	ccs_add_task_security(new_security, list);
+	return 0;
+}
+
 /**
  * ccs_find_task_security - Find "struct ccs_security" for given task.
  *
@@ -1316,7 +1306,7 @@ struct ccs_security *ccs_find_task_security(const struct task_struct *task)
 		smp_rmb();
 	rcu_read_lock();
 	list_for_each_entry_rcu(ptr, list, list) {
-		if (ptr->pid != task->pids[PIDTYPE_PID].pid)
+		if (ptr->task != task)
 			continue;
 		rcu_read_unlock();
 		/*
@@ -1344,33 +1334,14 @@ struct ccs_security *ccs_find_task_security(const struct task_struct *task)
 		if (task == current &&
 		    (ptr->ccs_flags & CCS_TASK_IS_IN_EXECVE) &&
 		    !current->in_execve) {
-			ccs_debug_trace("4");
+			ccs_debug_trace("1");
 			ccs_clear_execve(-1, ptr);
 		}
 		return ptr;
 	}
 	rcu_read_unlock();
-	if (task != current) {
-		/*
-		 * If a thread does nothing after fork(), caller will reach
-		 * here because "struct ccs_security" for that thread is not
-		 * yet allocated. But that thread is keeping a snapshot of
-		 * "struct ccs_security" taken as of ccs_task_create()
-		 * associated with that thread's "struct cred".
-		 *
-		 * Since that snapshot will be used as initial data when that
-		 * thread allocates "struct ccs_security" for that thread, we
-		 * can return that snapshot rather than &ccs_default_security.
-		 *
-		 * Since this function is called by only ccs_select_one() and
-		 * ccs_read_pid() (via ccs_task_domain() and ccs_task_flags()),
-		 * it is guaranteed that caller has called rcu_read_lock()
-		 * (via ccs_tasklist_lock()) before finding this thread and
-		 * this thread is valid. Therefore, we can do __task_cred(task)
-		 * like get_robust_list() does.
-		 */
-		return ccs_find_cred_security(__task_cred(task));
-	}
+	if (task != current)
+		return &ccs_default_security;
 	/* Use GFP_ATOMIC because caller may have called rcu_read_lock(). */
 	ptr = kzalloc(sizeof(*ptr), GFP_ATOMIC);
 	if (!ptr) {
@@ -1379,101 +1350,27 @@ struct ccs_security *ccs_find_task_security(const struct task_struct *task)
 		send_sig(SIGKILL, current, 0);
 		return &ccs_oom_security;
 	}
-	*ptr = *ccs_find_cred_security(task->cred);
-	/* We can shortcut because task == current. */
-	ptr->pid = get_pid(((struct task_struct *) task)->
-			   pids[PIDTYPE_PID].pid);
-	ptr->cred = NULL;
+	*ptr = ccs_default_security;
+	ptr->task = task;
 	ccs_add_task_security(ptr, list);
 	return ptr;
 }
 
 /**
- * ccs_copy_cred_security - Allocate memory for new credentials.
+ * __ccs_free_task_security - Release memory associated with "struct task_struct".
  *
- * @new: Pointer to "struct cred".
- * @old: Pointer to "struct cred".
- * @gfp: Memory allocation flags.
- *
- * Returns 0 on success, negative value otherwise.
- */
-static int ccs_copy_cred_security(const struct cred *new,
-				  const struct cred *old, gfp_t gfp)
-{
-	struct ccs_security *old_security = ccs_find_cred_security(old);
-	struct ccs_security *new_security =
-		kzalloc(sizeof(*new_security), gfp);
-	if (!new_security)
-		return -ENOMEM;
-	*new_security = *old_security;
-	new_security->cred = new;
-	ccs_add_cred_security(new_security);
-	return 0;
-}
-
-/**
- * ccs_find_cred_security - Find "struct ccs_security" for given credential.
- *
- * @cred: Pointer to "struct cred".
- *
- * Returns pointer to "struct ccs_security" on success, &ccs_default_security
- * otherwise.
- */
-static struct ccs_security *ccs_find_cred_security(const struct cred *cred)
-{
-	struct ccs_security *ptr;
-	struct list_head *list = &ccs_cred_security_list
-		[hash_ptr((void *) cred, CCS_TASK_SECURITY_HASH_BITS)];
-	rcu_read_lock();
-	list_for_each_entry_rcu(ptr, list, list) {
-		if (ptr->cred != cred)
-			continue;
-		rcu_read_unlock();
-		return ptr;
-	}
-	rcu_read_unlock();
-	return &ccs_default_security;
-}
-
-/**
- * ccs_task_security_gc - Do garbage collection for "struct task_struct".
+ * @task: Pointer to "struct task_struct".
  *
  * Returns nothing.
- *
- * Since security_task_free() is missing, I can't release memory associated
- * with "struct task_struct" when a task dies. Therefore, I hold a reference on
- * "struct pid" and runs garbage collection when associated
- * "struct task_struct" has gone.
  */
-static void ccs_task_security_gc(void)
+static void __ccs_free_task_security(const struct task_struct *task)
 {
-	static DEFINE_SPINLOCK(lock);
-	static atomic_t gc_counter = ATOMIC_INIT(0);
-	unsigned int idx;
-	/*
-	 * If some process is doing execve(), try to garbage collection now.
-	 * We should kfree() memory associated with "struct ccs_security"->ee
-	 * as soon as execve() has completed in order to compensate for lack of
-	 * security_bprm_free() and security_task_free() hooks.
-	 *
-	 * Otherwise, reduce frequency for performance reason.
-	 */
-	if (!atomic_read(&ccs_in_execve_tasks) &&
-	    atomic_inc_return(&gc_counter) < 1024)
+	unsigned long flags;
+	struct ccs_security *ptr = ccs_find_task_security(task);
+	if (ptr == &ccs_default_security || ptr == &ccs_oom_security)
 		return;
-	if (!spin_trylock(&lock))
-		return;
-	atomic_set(&gc_counter, 0);
-	rcu_read_lock();
-	for (idx = 0; idx < CCS_MAX_TASK_SECURITY_HASH; idx++) {
-		struct ccs_security *ptr;
-		struct list_head *list = &ccs_task_security_list[idx];
-		list_for_each_entry_rcu(ptr, list, list) {
-			if (pid_task(ptr->pid, PIDTYPE_PID))
-				continue;
-			ccs_del_security(ptr);
-		}
-	}
-	rcu_read_unlock();
-	spin_unlock(&lock);
+	spin_lock_irqsave(&ccs_task_security_list_lock, flags);
+	list_del_rcu(&ptr->list);
+	spin_unlock_irqrestore(&ccs_task_security_list_lock, flags);
+	kfree_rcu(ptr, rcu);
 }
